/* maf_tba -- check a maf-file for threadability in all sequences
*
*  Command syntax is:
*	maf_tba maf-file
*
*  Each sequence is verified to thread through blocks.  Specifically, starting
*  at position 0, there is a chain of blocks such that (relative to the fixed
*  sequence) when a block ends at one position, another block starts at the
*  following position; the last block ends at the last position in the sequence.
*  Any block must be contained in the chain for at least one species.
*/

#include "lib.h"
#include "maf.h"

struct interval {
	int beg, end;
} endpoints;

static int compar(struct interval *a, struct interval *b) {
	if (a->beg > b->beg)
		return 1;
	if (a->beg < b->beg)
		return -1;
	return 0;
}

int main(int argc, char **argv)
{
	struct mafFile *mf;
	struct mafAli *a, *A;
	struct mafComp *c, ***comps;
	char *source[10000];
	int sourceSize[10000], nseq, nalign, i, j, nint;
	struct interval *ints;

	argv0 = "maf_tba";
	if (argc != 2)
		fatal("arg = maf-file");
	mf = mafReadAll(argv[1], 1);
	A = mf->alignments;

	for (nalign = nseq = 0, a = A; a != NULL; a = a->next, ++nalign) {
		for (c = a->components; c != NULL; c = c->next) {
			if (c->strand != '+')
				fatalf("currently don't handle - strand");
			for (i = 0; i < nseq; ++i)
				if (same_string(source[i], c->src))
					break;
			if (i == nseq) {
				if (nseq == 10000)
					fatal("Too many sequence names");
				source[nseq] = copy_string(c->src);
				sourceSize[nseq] = c->srcSize;
				++nseq;
			}
		}
	}

	comps = ckalloc(nalign*sizeof(struct mafComp **));
	for (a = A, i = 0; i < nalign; ++i, a = a->next) {
		comps[i] = ckalloc(nseq*sizeof(struct mafComp *));
		for (j = 0; j < nseq; ++j)
			comps[i][j] = NULL;
		for (c = a->components; c != NULL; c = c->next) {
			for (j = 0; j < nseq; ++j)
				if (same_string(c->src, source[j]))
					break;
			if (j == nseq)
				fatalf("Didn't find %s", c->src);
			if (comps[i][j] != NULL)
				fatalf("Collision over %s", c->src);
			comps[i][j] = c;
		}
	}

	ints = ckalloc(nalign*sizeof(struct interval));
	for (j = 0; j < nseq; ++j) {
		for (nint = i = 0; i < nalign; ++i)
			if ((c = comps[i][j]) != NULL) {
				ints[nint].beg = c->start;
				ints[nint].end = c->start + c->size - 1;
				++nint;
			}
		qsort((void *) ints, (size_t)nint, sizeof(struct interval),
		  (const void *)compar);
		if (ints[0].beg != 0)
			fatalf("%s starts at %d", source[j], ints[0].beg);
		for (i = 1; i < nint; ++i)
			if (ints[i-1].end != ints[i].beg-1)
				fatalf("%s: confused around %d-%d", source[j],
				  ints[i-1].end, ints[i].beg);
		if ((i = ints[nint-1].end+1) != sourceSize[j])
			fatalf("%d reached %d instead of %d", source[j], i,
			  sourceSize[j]);
		printf("%s chains properly to %d\n", source[j], i);
	}

	printf("All blocks in %s chain properly.\n", argv[1]);
	return 0;
}
