// multiz.c -- align two reference-sequence alignsets, guided by a pairwise
// alignset of the two reference sequences.

#define VERSION 8

/* ------------------------ Input and Output -----------------------------------
*
*  In general, the command line names three alignment files of forms
*	(seq1+seqN).maf
*	(seq1+seq2+...+seq(N-1)).maf
*	(seqN+seq(N+1)+...+seqP).maf
*  and the output has the form:
*	(seq1+seq2+...seq(N-1)+seqN+...+seqP).maf
*  The first two input files and the output file (i.e., all of the
*  seq1-referenced alignsets) are ordered by increasing start position in seq1.
*
*  Multiz also supports simpler invocation formats, where either the second or
*  the third file-name argument (but not both) is replaced by "-".  If the
*  second file name is given as "-", multiz behaves is if there existed a
*  second-named file containing one-row alignments consisting of first rows
*  from the first-named alignset (with dashes removed).  If the third file name
*  is given as "-", multiz behaves as if there existed a third-named file
*  containing one-row alignments consisting of second rows of the first-named
*  alignset.
*
*  There are two optional arguments that can come before the three file-name
*  arguments.  One has the form M=n where n is a positive integer telling the
*  minimum width of an alignment that will be printed.  The other has the form
*  R=n where n is a non-negative integer giving the "radius" of the region
*  where dynamic programing is to be performed.  Thus, a command might be:
*	multiz M=8 R=30 hc.maf hmr.maf cpd.maf > hmrcpd.maf
*  In general, raising R may produce somewhat better alignment, while slowing
*  execution.  The default values of M and R, called MIN_WIDTH and DP_RADIUS
*  respectively, are defined below.
*
*  Multiz writes its output to stdout.  In addition to the new alignments that
*  it computes, multiz writes segments of alignments in the first two input
*  files (i.e., the seq1-referenced alignsets) that are not put into a
*  seq1+...+seqP alignment;  however, multiz discards unused portions of the
*  seqN+seq(N+1)+...+seqP alignments, in keeping with the reference-sequence
*  approach.
*
*  Input and output follows somewhat different rules when multiz is invoked by
*  the tba program.  The main difference is that there are two additional
*  command-line arguments, specifying where to write segments of alignments in
*  the second and third arguments that aren't put into a taller alignment.
*  The only alignments written to stdout are the ones created by dynamic
*  programming, and unused portions of the first input file are discarded.
*  Also, these additional file names can optionally be followed by the word
*  "nohead", asking that the maf-file header line not be written to stdout.
*  The final difference is that an M=n value is interpreted as the minimun
*  width of seq1+..+seqP alignment that will be written; arbitrarily small
*  unused portions of the first and second input files are reported.
* --------------------------------------------------------------------------- */


/* --------------------- Algorithm and Implementation --------------------------
*
*  Outline:  We find each combination of a seq1+seqN alignment in the first
*  file and a seq1+seq2+...+seq(N-1) alignment in the second file whose seq1
*  components intersect.  Let the intersection be positions Beg1-End1 of seq1,
*  which aligned to BegN-EndN in seqN by the first alignment.  We then look for
*  each seqN+seq(N+1)+...+seqP alignment in the third file whose seqN component
*  intersects BegN-EndN, say in begN-endN, which is the image of beg1-end1 of
*  seq1 under the first alignment.  We align the beg1-end1 segment of the
*  second alignment with the begN-endN segment of the third alignment (to
*  create a seq1+seq2+...+seqP alignment), using dynamic programming limited to
*  a region determined by the first alignment.
*
*  Multiz needs to deal with three input modes (depending on which input file,
*  if any, is specified as "-") and two output modes (one for stand-alone use
*  and another when multiz is called by tba).  Program logic is a bit
*  complicated.  Igoring output for a moment, input is processed as follows.
*  In the description, A1, A2 and A3 denote an arbitrary alignment from the
*  first, second or third alignment file, respectively, and a names of an
*  algorithmic step (e.g., NOFILE2) is also the name of the corresponding
*  procedures later in this file.
*
*  MAIN.  If the second alignment file is specified as "-", then do step
*	NOFILE2.  Otherwise, for each (A1,A2) pair that overlap in their
*	reference-sequence row do the following.  If the third sequence file
*	is specified as "-", then do step NOFILE3, and otherwise do step
*	OVERLAP.
*
*  NOFILE2.  For each A1, create a pseudo-A2 whose only row equals A1's first
*	row with dashes removed, and perform step OVERLAP for that (A1,A2) pair.
*
*  NOFILE3.  For a given (A1,A2) pair that overlap in their reference rows,
*  	create a pseudo-A3 whose only row is the second row of A1 with dashes
*	removed.  Apply step DO_YAMA to (A1,A2,A3).
*
*  OVERLAP.  For a given (A1,A2) pair that overlap in their reference rows, say
*	in columns c-d of A1, find all A3's whose reference row overlaps a
*	subset of columns c-d of A1's second row.  For each such (A1,A2,A3),
*	apply step DO_YAMA.
*
*  ALIGN.  For a given (A1,A2,A3) triple for which there is a slice (run of
*	consecutive columns) of A1 where the first row is a subrange of A2's
*	reference row and the second row is a subrange of A3's reference row,
*	align the corresponding slices of A2 and A3.  A1 is taken as showing an
*	approximation to the correct alignment.
*
*  Additional complexity in dealing with input arises because multiz handles
*  arbitrary reference-sequence alignments, not just "single-coverage"
*  alignments.  That is, though the first and second alignsets must be given in
*  non-decreasing order of starting position in the reference sequence, any of
*  the three input alignsets can contain alignments with overlapping reference
*  rows.  (To be completely honest, we permit a few alignments to be given out
*  of order, e.g., when blastz outputs from overlapping segments of the genome
*  are concatenated; if an alignment starts before the start of a preceding
*  alignment, we discard it.)  To handle alignments that overlap in their
*  reference-sequence row, we keeps lists of the currently-active alignments
*  from the first two input files.  When the sweep along their reference
*  sequence passes beyond one of these alignments, its unused segments can be
*  determined (see the next paragraph) and the alignment can be discarded.
*
*  Output.  Alignments obtained from an A2-slice and an A3-slice by step DO_YAMA
*  are written to stdout.  For every "real" alignment A1, A2, or A3 (i.e., not
*  the pseudo-alignments create for file names "-"), multiz keeps track of which
*  columns have been assembled into "taller" alignments by step DO_YAMA.  In
*  stand-alone mode (where no additional files are specified on the command
*  line), multiz also writes the unused slices of A1 and A2 to stdout (and
*  ignores the unused A3-slices).  When multiz is invoked by tba, two additional
*  files are specified; multiz writes the unused A2-slices to the first and the
*  unused A3-slices to the second (unused A1-slices are ignored).
*/

#include "lib.h"
#include "maf.h"
#include "mz_util.h"
#include "mz_scores.h"
#include "mz_preyama.h"

#define DP_RADIUS 10
#define MIN_WIDTH 0

static int radius=DP_RADIUS;
static int min_width=MIN_WIDTH;
static int lax=0;

static char *USAGE =
"args: [R=?] [M=?] (seq1+seqN).maf (seq1+seq2+...+seq(N-1)).maf (seqN+seq(N+1)+...+seqP).maf";

FILE *outfile2, *outfile3;

/* ---------------------- start of code to trim alignments ---------------------
*  We experimented with code to guarantee that terminal segments of a printed
*  alignment are guaranteed to have non-negative score.  Also, we wrote code
*  to produce maximal-scoring segments of the alignments produced by dynamic
*  programming.  However, for now, we're keeping things simple and just removing
*  rows that contain only dashes.
*
*  When multiz is called by tba, we need to keep every base-pair, but otherwise
*  (i.e., when outfile2 == NULL) we discard the alignment if the row for the
*  reference sequence is all dashes.
*/
static struct mafAli *mafTrim(struct mafAli *a) {
	struct mafComp *c, *prev, *cnext;

	if (a == NULL || (c = a->components) == NULL)
		return NULL;
	for (prev = NULL; c != NULL; c = cnext) {
		cnext = c->next;
		if (c->size <= 0) {
			if (prev == NULL) {
				if (outfile2 == NULL)
					return NULL;
				a->components = c->next;
			} else
				prev->next = c->next;
			mafCompFree(&c);
		} else
			prev = c;
	}
	a->score = mafScoreRange(a, 0, a->textSize);
	return a;
}
/* ----------------------- end of code to trim alignments ------------------- */

/* ------- idiosyncratic code to detect unused segments of alignments ------- */

// put unused segments of nodes in the output list (for stand-alone multiz use)
static struct mafAli *nodeUnused(struct node *n, struct mafAli *output_list) {
	struct mafAli *a = n->maf, *b;
	int i, j, sz;

	sz = a->textSize;
	for (j = 0; j < sz; ) {
		// set i to next unused position
		for (i = j; i < sz && n->used[i] == 1; ++i)
			;
		// set j just beyond last consecutive unused position
		for (j = i+1; j < sz && n->used[j] == 0; ++j)
			;
		if (i < sz && (b = mafTrim(mafSlice(a, i, j))) != NULL &&
		    b->textSize >= min_width)
			output_list = mafInsert(output_list, b);
	}
	return output_list;
}

// print unused segments of nodes on the designed stream (used by tba)
void nodeUnused2(struct node *n, FILE *fp) {
	struct mafAli *a = n->maf, *b;
	int i, j, sz;

	sz = a->textSize;
	for (j = 0; j < sz; ) {
		// set i to next unused position
		for (i = j; i < sz && n->used[i] == 1; ++i)
			;
		// set j just beyond last consecutive unused position
		for (j = i+1; j < sz && n->used[j] == 0; ++j)
			;
		if (i < sz && (b = mafTrim(mafSlice(a, i, j))) != NULL)
			mafWrite(fp, b);
	}
}
/* ----------- end of code to detect unused segments of alignments ---------- */

/* ---------- start of code to manage third input file of alignments -----------
*  Alignments in the third file have a different reference sequence than the
*  other two input files and the output file.  Their order and/or orientation,
*  as given, may need to be changed.  Thus, we read all of them at once.  With
*  each overlap between alignments in each of the first two files, we mark
*  all alignments in the third file as "untouched" (by setting score to 0.0),
*  then dish them out in the desired order and orientation.
*/

char dna_compl[256] = 
  "                                             -                  "
  " TVGH  CD  M KN   YSA BWXR       tvgh  cd  m kn   ysa bwxr      "
  "                                                                "
  "                                                                ";
/* .............................................-.................. */
/* @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~. */
/* ................................................................ */
/* ................................................................ */

static void do_revcompl(char *s, int len)
{
	char c, *p = s + len - 1;

	while (s<=p) {
		c = dna_compl[(uchar)*s]; 
		*s = dna_compl[(uchar)*p]; 
		*p = c;
		++s, --p;
	}
}

/* Implementation Note: When multiz is called by tba, the third alignment file
*  is totally ordered, so nextAlign3 can be made much more efficient in this
*  special case.  I profiled multiz on that kind of data and learned that
*  nextAlign3 was taking about 1% of the execution time.  Ergo: don't program
*  this "optimization".
*/

/* Find the active (i.e., score > 0.0) alignment from the specified chromosome
*  with the smallest end1 >= pos.  If no alignments on chr end at pos or later,
*  return NULL.
*/
static struct node *nextAlign3(struct node *A3, char *src, int pos,
  char strand) {
	struct node *n, *best = NULL;
	struct mafAli *a;
	struct mafComp *c;
	int end, best_end = -1;
	int i, j, len, used;

	for (n = A3; n != NULL; n = n->next) {
		a = n->maf;
		if (a->score == 0.0)	// marked as "used"
			continue;
		c = a->components;
		if (c->strand == strand)
			end = c->start + c->size - 1;
		else
			end = c->srcSize - c->start - 1;
		if (same_string(c->src, src) && end >= pos &&
		    (best == NULL || end < best_end)) {
			best = n;
			best_end = end;
		}
	}

	if (best == NULL)
		return NULL;

	a = best->maf;
	c = a->components;
	if (c->strand != strand) {
		for ( ; c; c = c->next) {
			c->start = c->srcSize - (c->start + c->size);
			c->strand = (c->strand == '-' ? '+' : '-');
			do_revcompl(c->text, a->textSize);
		}
	       
		len = best->maf->textSize;
                for (i=0, j=len-1; i<=j; i++, j--) {
                  used = best->used[i];
                  best->used[i]=best->used[j];
                  best->used[j]=used;
                }
		
	}
	return best;
}
/* ---------- end of code to manage third input file of alignments ---------- */

/* Given: a pairwise alignment, a1, and intervals beg1-end1 and begN-endN, where
*    (1) a1 aligns beg1 to begN and end1 to endN,
*    (2) positions beg1-end1 are contained in the top row of alignment a2,
*    (3) positions begN-endN are contained in the top row of alignment a3.
*  Align the corresponding slices of a2 and a3, by calling the procedure
*  pre_yama, which calls yama (yet another multiple aligner), which does the
*  hard work.
*/
static struct mafAli *ALIGN(struct node *A1, struct node *A2,
  struct node *A3, struct mafAli *output_list,
  int beg1, int end1, int begN, int endN){
	struct mafAli *a, *a1 = A1->maf, *a2 = A2->maf, *a3 = A3->maf;
	struct mafComp *c;
	int beg, end;

	if ((a = pre_yama(a1, a2, a3, beg1, end1, begN, endN, radius)) != NULL
	    && a->textSize >= min_width && (a = mafTrim(a)) != NULL) {
		output_list = mafInsert(output_list, a);
		// a tiny sanity check
		c = a->components;
		beg = c->start;
		end = beg + c->size - 1;
		if (beg != beg1 || end != end1)
		       fatalf("ALIGN: beg = %d, beg1 = %d, end = %d, end1 = %d",
			 beg, beg1, end, end1);
		// mark the parts of A1, A2 and A3 that were used 
		nodeMark(A1, beg, end);
		nodeMark(A2, beg, end);
		nodeMark(A3, begN, endN);
	}
	return output_list;
}

// trim the pairwise alignment further, until reaching columns without a dash
static void endsBlunt(struct mafComp *row1, struct mafComp *row2, int size,
  int *beg1, int *begN, int col_beg, int *end1, int *endN, int col_end) {
	int b1 = *beg1, bN = *begN, e1 = *end1, eN = *endN;

	while (row1->text[col_beg] == '-' || row2->text[col_beg] == '-') {
		if (col_beg >= size)
			fatal("endsBlunt: overflow of col_beg");
		if (row1->text[col_beg] != '-')
			++b1;
		if (row2->text[col_beg] != '-')
			++bN;
		++col_beg;
	}
	while (row1->text[col_end] == '-' || row2->text[col_end] == '-') {
		if (col_end < 0)
			fatal("endsBlunt: underflow of col_end");
		if (row1->text[col_end] != '-')
			--e1;
		if (row2->text[col_end] != '-')
			--eN;
		--col_end;
	}
	*beg1 = b1;
	*begN = bN;
	*end1 = e1;
	*endN = eN;
}

/* Given: a pairwise alignment, a1, and another alignment, a2, with the same
*  reference sequence.  Find the first and last positions in the first
*  alignment's two sequences, plus the columns in that alignment, corresponding
*  to the intersection of a1 and a2.
*/
static void endsOverlap(struct mafAli *a1, struct mafAli *a2,
  int *beg1, int *begN, int *col_beg, int *end1, int *endN, int *col_end) {
	struct mafComp *row1, *row2, *c2;
	int b1, bN, cb, e1, eN, ce, b2, e2;

	row1 = a1->components;
	row2 = row1->next;
	c2 = a2->components;
	b1 = row1->start;
	bN = row2->start;
	b2 = c2->start;
	cb = 0;
	while (b1 < b2) {
		if (cb >= a1->textSize)
			fatal("endsOverlap: overflow of col_beg");
		if (row1->text[cb] != '-')
			++b1;
		if (row2->text[cb] != '-')
			++bN;
		++cb;
	}
	e1 = row1->start + row1->size - 1;
	eN = row2->start + row2->size - 1;
	e2 = b2 + c2->size - 1;
	ce = a1->textSize - 1;
	while (e1 > e2) {
		if (ce < 0)
			fatal("endsOverlap: underflow of col_end");
		if (row1->text[ce] != '-')
			--e1;
		if (row2->text[ce] != '-')
			--eN;
		--ce;
	}
	*beg1 = b1;
	*begN = bN;
	*col_beg = cb;
	*end1 = e1;
	*endN = eN;
	*col_end = ce;
}

/* Given: a seq1+seqN alignment (a1) and a seq1+seq2+..+seq(N-1) alignment (a2)
*  that overlap at positions Beg1-End1 of seq1, corresponding to columns
*  Col_beg to Col_end of a1.  Find the seqN+seq(N+1)+...+seqP alignments (a3)
*  that overlap columns Col_beg-Col_end of a1 (in seqN).  Determine beg1,
*  begN, end1 and endN such that:
*	1. Beg1 <= beg1 <= end1 <= End1
*	2. a1 aligns position beg1 to begN (column col_beg) and end1 to endN
*	   (column col_end)
*	3. the first row of a3 includes positions begN-endN of seqN
*  If these slices (runs of columns) of a2 and a3, are non-empty, align them.
*/
static struct mafAli *OVERLAP(struct node *A1, struct node *A2,
  struct mafAli *output_list, struct node *A3){
	struct node *n3;
	struct mafAli *a1 = A1->maf, *a2 = A2->maf, *a3;
	struct mafComp *c, *c1, *c2;
	int i, Beg1, End1, BegN, EndN, beg1, end1, begN, endN, Col_beg,
	  Col_end, col_beg, col_end, beg3, end3;

	// Beg1 etc. (initial capital letter) refer to a1 and a2
	endsOverlap(a1, a2, &Beg1, &BegN, &Col_beg, &End1, &EndN, &Col_end);
	c = a1->components;
	c1 = c->next;
	c2 = a2->components;

	
	// mark each alignment in the third file as unused
	for (n3 = A3; n3 != NULL; n3 = n3->next)
		n3->maf->score = 1.0;
	// loop over overlapping entries of third alignment file
	for (i = BegN;
	     (n3 = nextAlign3(A3, c1->src, i, c1->strand)) != NULL &&
	      (a3 = n3->maf) != NULL &&
     	      (beg3 = a3->components->start) <= EndN; i = beg3) {
		a3->score = 0.0;	// mark as used
		end3 = beg3 + a3->components->size - 1;
		// beg1 etc (initial lower-case letter) are specific to a3

		// shrink ends in a1 to overlap with a3
		beg1 = Beg1;
		begN = BegN;
		col_beg = Col_beg;
		while (begN < beg3 && col_beg <= a1->textSize) {
			if (c->text[col_beg] != '-')
				++beg1;
			if (c1->text[col_beg] != '-')
				++begN;
			++col_beg;
		}
		if (col_beg >= a1->textSize)
			continue;

		end1 = End1;
		endN = EndN;
		col_end = Col_end;
		while (endN > end3 && col_end >= 0) {
			if (c->text[col_end] != '-')
				--end1;
			if (c1->text[col_end] != '-')
				--endN;
			--col_end;
		}
		if (col_end < 0)
			continue;
		endsBlunt(c, c1, a1->textSize, &beg1, &begN, col_beg, &end1,
		  &endN, col_end);
		if (beg1 > end1 || begN > endN)
			continue;
		output_list =
		  ALIGN(A1, A2, n3, output_list, beg1, end1, begN, endN);
	}
	return output_list;
}

// Build a one-row alignment (the 2nd or 3rd file-name is "-")
static struct mafAli *mafOneRow(struct mafComp *c, int size) {
	struct mafAli *a = ckalloc(sizeof(struct mafAli));
	struct mafComp *newc;
	int i, j;
	uchar x;

	a->components = newc =
	   mafNewComp(c->src, c->start, c->size, c->strand, c->srcSize, size);
	for (j = i = 0; i < size; ++i)
		if ((x = c->text[i]) != '-')
			newc->text[j++] = x;
	newc->text[j] = '\0';
	a->textSize = j;
	return a;
}

/* The third command-line argment is '-'.  For a given overlapping pair of
*  alignments, one from the first file and one from the second, we build a
*  one-component mafAli whose row equals the first alignment's second row with
*  dashes removed.  This mafAli serves as the only alignment in the third file
*  for this specific overlapping pair.
*/
static struct mafAli *NOFILE3(struct node *A1, struct node *A2,
  struct mafAli *output_list) {
	struct node *n3;
	struct mafAli *a1 = A1->maf, *a2 = A2->maf;
	struct mafComp *c, *c1;
	int beg1, end1, begN, endN, col_beg, col_end;

	c = a1->components;
	c1 = c->next;
	endsOverlap(a1, a2, &beg1, &begN, &col_beg, &end1, &endN, &col_end);
	endsBlunt(c, c1, a1->textSize, &beg1, &begN, col_beg, &end1, &endN,
	  col_end);
	if (beg1 <= end1 && begN <= endN) {
		n3 = nodeNew(mafOneRow(c1, a1->textSize));
		output_list =
		   ALIGN(A1, A2, n3, output_list, beg1, end1, begN, endN);
		nodeFree(&n3);
	}
	return output_list;
}

// Get next maf entry from first or second alignment file.
// Discard out-of-order mafs.
static struct mafAli *monoNext(struct mafFile *mf, char *filename, int old_beg)
{
	struct mafAli *a;
	struct mafComp *c;

	for ( ; ; ) {
		if ((a = mafNext(mf)) == NULL)
			break;
		if ((c = a->components) == NULL)
			fatalf("maf entry in %s has no components", filename);
		if (c->start < old_beg && lax == 0)
			fatalf("maf starting at %d in %s is out of order",
			  c->start, filename);
		if (c->start >= old_beg)
			break;
		if ((c = c->next) == NULL)
			fatalf("maf entry in %s has only one component",
			  filename);
	}
	return a;
}

/* If the second command-line argment is '-',  then for each alignment in the
*  first file, we build a one-component maf whose row equals that alignment's
*  first row with dashes removed, which serves the role of the only overlapping
*  alignment in the second file.
*/
static void NOFILE2(struct mafFile *mf1, char *filename, struct node *A3) {
	struct mafAli *a1, *a2, *output_list = NULL;
	struct mafComp *c, *c1;
	struct node *A1, *A2;
	int beg1;

	beg1 = 0;
	while ((a1 = monoNext(mf1, filename, beg1)) != NULL) {
		c = a1->components;
		beg1 = c->start;
		c1 = c->next;
		A1 = nodeNew(a1);
		a2 = mafOneRow(c, a1->textSize);
		A2 = nodeNew(a2);
		output_list = mafFlush(output_list, beg1);
		output_list = OVERLAP(A1, A2, output_list, A3);
		output_list = nodeUnused(A1, output_list);
		nodeFree(&A1);
		nodeFree(&A2);
	}
	
	(void) mafFlush(output_list, INT_MAX);
}

// MAIN()
int main(int argc, char **argv) {
	struct mafFile *mf1, *mf2, *mf3;
	struct mafAli *a1, *a2, *a3, *output_list;
	struct node *A1, *A2, *A3, *A1_first, *A1_last, *Atmp, *A3_last;
	struct mafComp *c1, *c2;
	int i, x, beg0, beg1, beg2, end1, end2, last_beg2, last_beg1, mf1_eof;
	char cmd[30];

	sprintf(cmd, "multiz.v%d", VERSION);
	argv0 = cmd;

	if (argc == 2 && same_string(argv[1], "version")) {
		printf("# %s\n", cmd);
		exit(0);
	}
	if (same_string(argv[argc-1], "lax")) {
		lax = 1;
		--argc;
	}
	while (argc > 1 && strchr("RM", (x=argv[1][0])) && argv[1][1] == '=') {
		if (x == 'R') {
			radius = atoi(argv[1]+2);
			if (radius < 0)
				fatal("radius cannot be negative");
			//fprintf(stderr, "radius changed to %d\n", radius);
		} else if (x == 'M') {
			min_width = atoi(argv[1]+2);
			if (min_width < 0)
				fatal("min_width cannot be negative");
			//fprintf(stderr, "min_width set to %d\n",  min_width);
		} else
			fatalf("illegal flag %c", x);
		++argv;
		--argc;
	}

	if (argc != 4 && argc != 6 && argc != 7)
		fatal(USAGE);

	if (argc == 7 && !same_string(argv[6], "nohead"))
		fatal("seventh argument must be 'nohead'");

	init_scores70();

	if (argc != 7)
		mafWriteStart(stdout, cmd);
	if (argc == 4) {
		// not called by tba
		printf("# %s R=%d M=%d", cmd, radius, min_width);
		for (i = 1; i < argc; ++i)
			printf(" %s", argv[i]);
		if (lax == 1)
			printf(" lax");
		putchar('\n');
	}

	if (argc >= 6) {
		outfile2 = ckopen(argv[4], "w");
		outfile3 = ckopen(argv[5], "w");
	} else
		outfile2 = outfile3 = NULL;
	mf1 = mafOpen(argv[1], 1);
	A3 = NULL;
	mf3 = NULL;
	if (!same_string(argv[3], "-")) {
		mf3 = mafReadAll(argv[3], 1);
		A3_last = NULL;
		for (a3 = mf3->alignments; a3 != NULL; a3 = a3->next) {
			Atmp = nodeNew(a3);
			if (A3_last == NULL)
				A3 = Atmp;
			else
				A3_last->next = Atmp;
			A3_last = Atmp;
		}
	}
	if (same_string(argv[2],"-")) {
		if (mf3 == NULL)
			fatal("cannot have two '-' arguments");
		NOFILE2(mf1, argv[1], A3);
		mafFileFree(&mf1);
		if (outfile3 != NULL)
			for ( ; A3 != NULL; A3 = A3->next)
				nodeUnused2(A3, outfile3);
		mafFileFree(&mf3);
		return 0;
	}
	/* The remainder of this procedure walks over every pair of alignments,
	*  one from file 1 and one from file 2, that overlap in their reference
	*  sequence (i.e., top row).  Files placed on the output list are
	*  written as soon as we can be certain about their order.
	*/
	mf2 = mafOpen(argv[2], 1);
	beg0 = beg1 = end1 = beg2 = end2 = last_beg2 = 0;
	last_beg1 = mf1_eof = 0;
	A1_first = A1_last = NULL;

	output_list = NULL;
	// for each seq1+seq2+...+seq(N-1) alignment do:
	while ((a2 = monoNext(mf2, argv[2], beg2)) != NULL) {
		c2 = a2->components;
		beg2 = c2->start;
		end2 = beg2 + c2->size - 1;
		A2 = nodeNew(a2);
		// flush expired mafs from the start of the seq1+seqN list
		// make the seq1+seqN list long enough
		while (A1_first != NULL) {
			c1 = A1_first->maf->components;
			if (c1->start + c1->size - 1 >= beg2)
				break;
			// seq1+seqN maf ends before the start of the current
			// seq1+seq2+...+seq(N-1) maf;
			// insert unused parts into output list and discard it
			Atmp = A1_first;
			if ((A1_first = A1_first->next) == NULL)
				A1_last = NULL;
			if (argc == 4)
				output_list = nodeUnused(Atmp, output_list);
			nodeFree(&Atmp);
		}
		while ((A1_last == NULL || last_beg1 <= end2) && !mf1_eof) {
			if ((a1 = monoNext(mf1, argv[1], last_beg1))
			  == NULL) {
				mf1_eof = 1;
				break;
			}
			A1 = nodeNew(a1);
			if (A1_last == NULL)
				A1_first = A1;
			else
				A1_last->next = A1;
			A1_last = A1;
			last_beg1 = a1->components->start;
		}
		// flush the output list
		x = beg2;
		if (A1_first != NULL) {
			c1 = A1_first->maf->components;
			x = MIN(x, c1->start);
		}
		output_list = mafFlush(output_list, x);
		// process the intersections
		for (A1 = A1_first; A1 != NULL; A1 = A1->next) {
			c1 = A1->maf->components;
			if (!same_string(c1->src, c2->src))
				fatalf("%s is reference in %s, %s in %s",
				  c1->src, argv[1], c2->src, argv[2]);
			beg1 = c1->start;
			end1 = beg1 + c1->size - 1;
			if (MAX(beg1, beg2) <= MIN(end1, end2))
				output_list = (mf3 == NULL ?
					NOFILE3(A1, A2, output_list) :
					OVERLAP(A1, A2, output_list, A3));
		}
		// insert the unused parts of a2 into output list
		if (outfile2 != NULL)
			nodeUnused2(A2, outfile2);
		else
			output_list = nodeUnused(A2, output_list);
		nodeFree(&A2);
	}
	while (A1_first != NULL) {
		if (argc == 4)
			output_list = nodeUnused(A1_first, output_list);
		Atmp = A1_first;
	       	A1_first = A1_first->next;
		nodeFree(&Atmp);
	}
	(void) mafFlush(output_list, INT_MAX);
	mafFileFree(&mf1);
	mafFileFree(&mf2);
	if (outfile3 != NULL)
		for ( ; A3 != NULL; A3 = A3->next)
			nodeUnused2(A3, outfile3);

	return 0;
}
