// mz_preyama.c -- prepare alignment data for mz_yama.c

#include "lib.h"
#include "maf.h"
#include "mz_util.h"
#include "mz_preyama.h"
#include "mz_yama.h"

// verify equality of characters in what are presumed to be identical sequences.
static void must_equal(char x, char y, int pos, // starting position
  struct mafComp *c1, int cbeg1, int cend1,	// begin and end columns
  struct mafComp *c2, int cbeg2, int cend2,
  int file2) {					// 2 or 3
	int i;
	char z;

	if (toupper(x) != toupper(y)) {
		fprintf(stderr, "%c != %c\n", x, y);
		fprintf(stderr, "in file 1, positions %d... of %s are:\n  ",
		  pos, c1->src);
		for (i = cbeg1; i <= cend1; ++i)
			if ((z = c1->text[i]) != '-')
				fputc(z, stderr);
		fputc('\n', stderr);
		fprintf(stderr, "while in file %d they are:\n  ", file2);
		for (i = cbeg2; i <= cend2; ++i)
			if ((z = c2->text[i]) != '-')
				fputc(z, stderr);
		fputc('\n', stderr);
		exit(1);
	}
}

// Convert an internal form of a multiple alignment to a maf.
// The info (src, etc.) comes from two other mafs.
static struct mafAli *mafBuild(uchar **A_new, int nrow, int ncol,
  struct mafAli *a2, int cbeg2, struct mafAli *a3, int cbeg3) {
	struct mafAli *A = ckalloc(sizeof(struct mafAli));
	struct mafComp *nc, *d, *prev_nc = NULL;
	int i, j, beg, strt;

	A->textSize = ncol;
	A->next = NULL;
	d = a2->components;
	beg = cbeg2;
	for (i = 0; i < nrow; ++i, d = d->next) {
		if (d == NULL) {
			d = a3->components;
			beg = cbeg3;
		}
		for (strt = d->start, j = 0; j < beg; ++j)
			if (d->text[j] != '-')
				++strt;
		nc = mafNewComp(d->src, strt, 0, d->strand, d->srcSize, ncol);
		if (prev_nc == NULL)
			A->components = nc;
		else
			prev_nc->next = nc;
		prev_nc = nc;
		for (j = 0; j < ncol; ++j)
			if ((nc->text[j] = A_new[j+1][i]) != '-')
				++(nc->size);
		nc->text[ncol] = '\0';
	}
	return A;
}

// Entry point: (documentation in mz_preyama.h)
struct mafAli *pre_yama(struct mafAli *a1, struct mafAli *a2,
  struct mafAli *a3, int beg1, int end1, int begN, int endN, int radius) {
	struct mafAli *val;
	struct mafComp *c, *c1, *c2, *c3, *d;
	int cbeg1, cend1, cbeg2, cend2, cbeg3, cend3, i, j, k, radi,
	  M, N, K, L, M_new, *LB, *RB;
	uchar **A, **B, **AL_new, x, y, z;

//printf("pre_yama: radius = %d\n", radius);
//mafWrite(stdout,a1);
//printf(" %d-%d\n", beg1, end1); mafWrite(stdout,a2);
//printf(" %d-%d\n", begN, endN); mafWrite(stdout,a3);
//fflush(stdout);

	if ((c = a1->components) == NULL || (c1 = c->next) == NULL)
		fatal("pre_yama: cannot find c and c1");
	if (c1->next != NULL)
		fatal("pre_yama: a1 is not a pairwise alignment");
	// count K = number of rows in a2
	for (d = c2 = a2->components, K = 0; d != NULL; ++K, d = d->next)
		;
	// count L = number of rows in a3
	for (d = c3 = a3->components, L = 0; d != NULL; ++L, d = d->next)
		;
	if (K == 0 || L == 0)
		fatal("pre_yama: an alignment has 0 rows");
	if (!same_string(c->src, c2->src))
		fatalf("pre_yama: first rows for sequences %s != %s",
		  c->src, c2->src);
	if (c->srcSize != c2->srcSize)
		fatalf("pre_yama: first row srcSizes %d != %d",
		  c->srcSize, c1->srcSize);
	if (c->strand != c2->strand)
		fatalf("pre_yama: first rows on opposite strands");
	if (!same_string(c1->src, c3->src))
		fatalf("pre_yama: first rows (2) for sequences %s != %s",
		  c1->src, c3->src);
	if (c1->srcSize != c3->srcSize)
		fatalf("pre_yama: first row (2) srcSizes %d != %d",
		  c1->srcSize, c3->srcSize);
	if (c1->strand != c3->strand)
		fatalf("pre_yama: first rows (2) on opposite strands");

	cbeg1 = mafPos2Col(c, beg1, a1->textSize);
	cend1 = mafPos2Col(c, end1, a1->textSize);
	// check that a1 aligns beg1 to end1, and begN to endN
	if (cbeg1 != mafPos2Col(c1, begN, a1->textSize))
		fatalf("pre_yama: mismatch of beg1 and begN");
	if (cend1 != mafPos2Col(c1, endN, a1->textSize))
		fatalf("pre_yama: mismatch of end1 and endN");
	cbeg2 = mafPos2Col(c2, beg1, a2->textSize);
	cend2 = mafPos2Col(c2, end1, a2->textSize);
	cbeg3 = mafPos2Col(c3, begN, a3->textSize);
	cend3 = mafPos2Col(c3, endN, a3->textSize);

	// determine and check sizes of slices of a2 and a3
	M = cend2 - cbeg2 + 1;
	N = cend3 - cbeg3 + 1;
	if (MAX(M, N) < 2)
		return NULL;

	// prepare first alignment for yama
	A = (uchar **)ckalloc(M*sizeof(uchar *)) - 1;
	A[1] = ckalloc(K*M*sizeof(uchar));
	for (i = 2; i <= M; ++i)
		A[i] = A[i-1] + K;
	for (i = 1; i <= M; ++i)
		for (j = 0, d = c2; j < K; ++j, d = d->next)
			A[i][j] = d->text[cbeg2 + i - 1];

	// prepare second alignment for yama
	B = (uchar **)ckalloc(N*sizeof(uchar *)) - 1;
	B[1] = ckalloc(L*N*sizeof(uchar));
	for (i = 2; i <= N; ++i)
		B[i] = B[i-1] + L;
	for (i = 1; i <= N; ++i)
		for (j = 0, d = c3; j < L; ++j, d = d->next)
			B[i][j] = d->text[cbeg3 + i - 1];

	// Determine sausage
	LB = ckalloc((M+1)*sizeof(int));
	RB = ckalloc((M+1)*sizeof(int));
	for (i = 0; i <= M; ++i) {
		LB[i] = 0;
		RB[i] = N;
	}

	// Column k of c and c1 corresponds to column i of A and column j of B.
	// Several "reality checks" guard against data/internal errors.
	for (i = j = 0, k = cbeg1; k <= cend1; ++k) {
		if ((x = c->text[k]) != '-') {
			do {
				if (++i > M)
					fatal("pre_yama: bad scene");
			} while ((z = A[i][0]) == '-');
#ifdef RM_HACK
			// RepeatMasker .align files sometimes give incorrect
			// sequence positions
			if (toupper(x) != toupper(z))
				return NULL;
#else
			must_equal(x, z, beg1, c, cbeg1, cend1,
					       c2, cbeg2, cend2, 2);
#endif
		}
		if ((y = c1->text[k]) != '-') {
			do {
				if (++j > N)
					fatal("pre_yama: ouch");
			} while ((z = B[j][0]) == '-');
			must_equal(y, z, begN, c1, cbeg1, cend1,
					       c3, cbeg3, cend3, 3);
		}
		// For fixed i, LB[i] and RB[i] are the minimum and maximum
		// corresponding j, resp.
		if (LB[i] == 0)
			LB[i] = j;
		RB[i] = j;
	}
	if (i != M || j != N)
		fatalf("pre_yama: i = %d, M = %d, j = %d, N = %d", i, M, j, N);
	// make LB and RB monotone
	for (i = j = 0; i <= M; ++i)
		LB[i] = j = MAX(j, LB[i]);
	for (i = M, j = N; i >= 0; --i)
		RB[i] = j = MIN(j, RB[i]);

	// expand line into a sausage
	radi = MIN(M, radius);
	for (i = M; i > radi; --i)
		LB[i] = MIN(MAX(LB[i]-radi,0),LB[i-radi]);
	for ( ; i >= 0; --i)
		LB[i] = 0;
	for (i = 0; i < M - radi; ++i)
		RB[i] = MAX(MIN(RB[i]+radi,N),RB[i+radi]);
	for ( ; i <= M; ++i)
		RB[i] = N;

	yama(A, K, M, B, L, N, LB, RB, &AL_new, &M_new);

	val = mafBuild(AL_new, K+L, M_new, a2, cbeg2, a3, cbeg3);

	free(AL_new[1]); free(AL_new+1);
	free(A[1]); free(A+1);
	free(B[1]); free(B+1);
	free(LB); free(RB);
//printf("returning:\n");
//mafWrite(stdout, val);
//fflush(stdout);
	return val;
}
